/*
 * ASCD : copyright (c) Aley Keprt 1998-2001
 * SimCoupe : Copyright (c) Allan Skillman 1996.
 *
 * File processor.c : main z80 loop and interrupt routines
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */


#include <stdio.h>
#include <signal.h>
#include <time.h>
#include <sys/time.h>
#include <sys/types.h>

#include "z80.h"
#include "mem.h"
#include "resource.h"
#include "util.h"
#include "mouse.h"
#include "intrface.h"
#include "video.h"
#include "audio.h"


#define af ((a<<8)|f)
#define bc ((b<<8)|c)
#define de ((d<<8)|e)
#define hl ((h<<8)|l)
#define bc_prime ((b1<<8)|c1)
#define af_prime ((a1<<8)|f1)
#define de_prime ((d1<<8)|e1)
#define hl_prime ((h1<<8)|l1)


//Z80 CPU registers
static uns16 pc, ix, iy, sp;
static uns8 a, b, c, f, d, e, h, l;
static uns8 r, a1, f1, b1, c1, d1, e1, h1, l1, i, iff1, iff2, im;


#define parity(a) (partable[a])


uns8 partable[256]={
  4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
    4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
    4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    0, 4, 4, 0, 4, 0, 0, 4, 4, 0, 0, 4, 0, 4, 4, 0,
    4, 0, 0, 4, 0, 4, 4, 0, 0, 4, 4, 0, 4, 0, 0, 4
};


#ifdef DISASS
#include "debug.h"
int disass,disass1;
#endif


int toplines;   //PAL lines in top border
int totallines; //PAL total lines


long n_frames;                  /* no. of interrupts since powerup */

int LineFreq;                   /* No of tstates / line generated */
int LineNo;                     /* Scan line being generated */
int InterruptType;              /* Type of z80 interupt to deal with */

static int audiomute;

static int IntActiveTime;       /* Time (in tstates) an int is active (Port 249 <> 0xff) */
static int MouseStrobeTime;     /* Time (in tstates) from a mouse strobe to a counter reset */

static int DelayedEI;           /* Flag and counter to carry out a delayed */

int flash_phase;                /* gets toggled between 1 and 0 twice per second */

int LineCycleCounter;           /* Counter used to count the cycles/line */


/* Syncronizing timer definitions */
/* SyncTimer MUST be declared volatile to prevent the optimizer
   Changing the Sync wait loop into an infinite loop */
volatile int SyncTimer;
static void emulator_sync_timer(int sigparam);
static void emulator_cancel_sync_timer(void);

static struct itimerval TheSyncTimerVal;


//-----------------------------------------------------------------------


/* from events.c */
extern void SamKeyboardHandler(int,int);

/* from samio.c */
extern void check_events(void);

/* from mouse.c */
extern int MouseStrobed;


//-----------------------------------------------------------------------


void set_registers(Z80_Regs *rr) {
  //
  //set all CPU registers
  //
  ix=rr->ix;
  iy=rr->iy;
  sp=rr->sp;
  a=rr->a;
  f=rr->f;
  b=rr->b;
  c=rr->c;
  d=rr->d;
  e=rr->e;
  h=rr->h;
  l=rr->l;
  r=rr->r;
  a1=rr->a1;
  f1=rr->f1;
  b1=rr->b1;
  c1=rr->c1;
  d1=rr->d1;
  e1=rr->e1;
  h1=rr->h1;
  l1=rr->l1;
  i=rr->i;
  im=rr->im;
  
  pc=rr->pc;
  iff1=rr->iff1;
  iff2=rr->iff2;
}


//-----------------------------------------------------------------------


void get_registers(Z80_Regs *rr) {
  //
  //get all CPU registers
  //
  rr->ix=ix;
  rr->iy=iy;
  rr->sp=sp;
  rr->a=a;
  rr->f=f;
  rr->b=b;
  rr->c=c;
  rr->d=d;
  rr->e=e;
  rr->h=h;
  rr->l=l;
  rr->r=r;
  rr->a1=a1;
  rr->f1=f1;
  rr->b1=b1;
  rr->c1=c1;
  rr->d1=d1;
  rr->e1=e1;
  rr->h1=h1;
  rr->l1=l1;
  rr->i=i;
  rr->im=im;
  
  rr->pc=pc;
  rr->iff1=iff1;
  rr->iff2=iff2;
}


//-----------------------------------------------------------------------


static int last_zxmode=-2;


void set_cpu_speed()
//
//update CPU speed constants (Sam/ZX mode)
//
{
  if(GetCfg(zxmode)==last_zxmode) return;
  last_zxmode=GetCfg(zxmode);
  
  /* Define the number of tstates/line generated */
  switch(GetCfg(zxmode)) {
  case 0: LineFreq=TSTATES_PER_LINE; break;
  case 1: LineFreq=TSTATES_PER_LINE_ZX; break;
  case 2: LineFreq=TSTATES_PER_LINE_128; break;
  }
  
  /* Calculate the number of tstates in the 100usecs an interrupt remains active */
  IntActiveTime   = 128; //30*LineFreq/64;
  MouseStrobeTime = LineFreq*100/64;  //LineFreq is 64microsec (PAL)
  
  //  simcoupe_mesg(SIMCOUPE_INFO,"LineFreq        : %d tstates", LineFreq);
  //  simcoupe_mesg(SIMCOUPE_INFO,"IntActiveTime   : %d tstates", IntActiveTime);
  //  simcoupe_mesg(SIMCOUPE_INFO,"MouseStrobeTime : %d tstates", MouseStrobeTime);
}


//-----------------------------------------------------------------------


void emulator_main_loop(void) {
  //
  //the main Z80 emulation loop
  //
  register unsigned long tstates=0;
  register unsigned int radjust=0;
  register uns8 ixoriy=0,new_ixoriy=0;
  register uns8 op;
  
  register int IntActiveCounter=0;   /* Counter used for int Active timing */
  register int MouseActiveCounter=0; /* Counter used for mouse Active timing */

  int refresh_interval;

  //-------------------------------------

  SyncTimer=!GetCfg(sync);
  
  reinit_sync_timer();
  
  set_cpu_speed();
  
  LineNo=totallines-1;
  
#ifdef USE_BORDER_EFFECTS
  BorderChangeInInstruction = False;
#endif
  while(1) {

#ifdef DISASS
    if(disass) {
      /*                      fprintf(stderr, "LineCycleCounter = %d, %d\n",LineCycleCounter,LineNo);*/
      disass=run_debug(pc);
      if (disass1 == 2) {
        fprintf(stderr, "Z80 CPU State\n-------------\n");
        fprintf(stderr, "  AF  = %04X\t\t  AF' = %04X\t\t  A = %02X  F = %02X\n",
          af, af_prime, a, f);
        fprintf(stderr, "  BC  = %04X\t\t  BC' = %04X\t\t  B = %02X  C = %02X\n",
          bc, bc_prime, b, c);
        fprintf(stderr, "  DE  = %04X\t\t  DE' = %04X\t\t  D = %02X  E = %02X\n",
          de, de_prime, d, e);
        fprintf(stderr, "  HL  = %04X\t\t  HL' = %04X\t\t  H = %02X  L = %02X\n",
          hl, hl_prime, h, l);
        fprintf(stderr, "  IX  = %04X\t\t  IY  = %04X\n\n",
          ix, iy);
        
        fprintf(stderr, "  PC  = %04X\t\t  SP  = %04X\n",
          pc, sp);
        
        fprintf(stderr, "   R  = %02X\t\t   I  = %02X\n",
          r, i);
        
        fprintf(stderr, "  interrupts %s, IM%d\n",
          iff1 ? "enabled" : "disabled", im);
        
        fprintf(stderr, "  FLAGS: C=%d N=%d P/V=%d H=%d Z=%d S=%d\n\n",
          f & F_CARRY    ? 1 : 0,
          f & F_NADD     ? 1 : 0,
          f & F_OVERFLOW ? 1 : 0,
          f & F_HCARRY   ? 1 : 0,
          f & F_ZERO     ? 1 : 0,
          f & F_NEG      ? 1 : 0);
      }
    }
#endif

    ixoriy=new_ixoriy;
    new_ixoriy=0;
/*
    //zjistujeme, na kterych adresach je aktivovano nahravani z kazety
    {
     static int load_zone;

     int new_load_zone= pc>=0x556 && pc<=0x604;
     if(new_load_zone!=load_zone) {
       if(!load_zone && pc>0x562) printf("entering load_zone: pc=0x%x\n",pc);
       load_zone^=1;
     }

     extern int SpectrumKeyboard;
     if(pc==0xe60e) {
       load_zone^=1;
     }
    }
*/
    DelayedEI=0;
    op=read_byte(pc);
    pc++;
    radjust++;

    switch(op) {
#include "z80ops.h"
    }
    
    LineCycleCounter+=tstates;

    if(new_ixoriy) continue;

    //if we have strobed the mouse, we will need to do a reset after 100us
    if(MouseStrobed && (MouseActiveCounter+=tstates)>=MouseStrobeTime) {
      MouseActiveCounter-=MouseStrobeTime;
      MouseStrobed=0;
      sam_mouse_reset(0);
    }
    
    //call the interrupt if required
    if(status_reg!=0xff) {

      //reset interrupt flag after IntActiveTime tstates
      IntActiveCounter+=tstates;
      if(IntActiveCounter >= IntActiveTime) {
        IntActiveCounter-=IntActiveTime;
        status_reg=0xff;
      }

      //maskable interrupts
      if(iff1 && !DelayedEI) {
        uns16 addr;

        if(op==0x76) pc++;
        iff1=iff2=0;

        LineCycleCounter+=8; /* accompanied by an input from the data bus */

        switch(im) {
        case 0: /* IM 0 */
        case 1: /* undocumented */
        case 2: /* IM 1 */
          /* there is little to distinguish between these cases */
          push2(pc);
          pc=0x38;
          break;
          
        case 3: /* IM 2 */
          addr=read_word((i<<8)|0xff);
          push2(pc);
          pc=addr;
          break;
        }
      }
    }

    /* Deal with the display */
    if(LineCycleCounter>=LineFreq) {

      //update screen
      if(LineNo<toplines) {

        //top border
        if(LineNo>=toplines-b_lines_drawn) borderline_update();

        //line interrupt (on next line, i.e. first paper line)
        if(line_int==0 && LineNo+1==toplines) {
          IntActiveCounter=0;
          status_reg=0xfe;
        }


      } else if(LineNo<toplines+192) {

        //paper
        line_update(LineNo-toplines);
        attr_reg=0;

        //line interrupt (on next line)
        if(line_int<192 && LineNo+1==toplines+line_int) {
          IntActiveCounter=0;
          status_reg=0xfe;
        }

      } else {
        
        //bottom border
        if(LineNo<toplines+192+b_lines_drawn) borderline_update();
        attr_reg=255;
      }
      
      //move to next line!
      LineNo++;
      LineCycleCounter-=LineFreq;
      BeeperNextLine();

      //flyback?
      if(LineNo>=totallines) {

        //frame interrupt
        IntActiveCounter=0;
        status_reg=0xf7;

        n_frames++;
        
        check_events();
        set_cpu_speed();
        
        LineNo=0;
        
        wait_for_timer();
        
        FlyBack();
        
        /* every 25 frames -> twice/sec. */
        if(n_frames%25==0 && cfg.flash && VMPR_SCREEN_MODE<MODE_3) {
          /* Sam modes 1 and 2 use the flashing attribute bit. Every frame
             the flash_phase is flipped (pen <-> paper in the screen write
             routines) and ClutChangeInLine is set to the very last display
             line to make sure the whole display is updated */
          flash_phase^=1;
          ClutChangeInLine=totallines;
        }
        
        /* Deal with NMI type interrupts */
        switch(InterruptType) {
        case Z80_none:
          break;
          
        case Z80_nmi:
          if(read_byte(pc)==0x76) pc++;
          iff2=iff1;
          iff1=0;
          /* The Z80 performs a machine fetch cycle for 5 Tstates
          but ignores the result.  It takes a further 10 Tstates
          to jump to the NMI routine at 0x66. */
          LineCycleCounter+=16;
          push2(pc);
          pc=0x66;
          InterruptType=Z80_none;
          break;
          
        case Z80_reset:
          a=f=b=c=d=e=h=l=a1=f1=b1=c1=d1=e1=
            h1=l1=i=r=iff1=iff2=im=0;
          ix=iy=sp=pc=0;
          radjust=0;
          init_ports();
          set_emulation_mode(cfg.zxmode);
          InterruptType=Z80_none;

          //by Aley Keprt 1999,2001
          //reset SAA1099
          {
            int i;
            for(i=31;i>=0;i--) {
              saa1099_control_port_w(0,0,i);
              saa1099_write_port_w(0,0,0);
            }
          }
          break;
          
          /* Special 'interrupt' type which stop and restart */
          /* the processor for UI routines                   */
        case Z80_menu:
          
          /**************************/
          /* halt emulation - begin */
          
          /* Temporarily mute sound */
          AudioMute();
          audiomute=1;

          InterruptType=Z80_none;
          BorderChangeInLine=totallines;
          
#ifdef INT_TIMER
          TimerQuit();
#endif
          KeyboardClose();
          
          {
            int old_scale=VideoGetScale();

          /* halt emulation - end */
          /************************/

            ui_run();
          
          /******************************/
          /* continue emulation - begin */
          
            KeyboardInit(SamKeyboardHandler);

            VideoSetScale(old_scale);
            /* Call vmpr to reset the screen write pointers
               and fetch the screen from memory */
            out_vmpr(vmpr);
          }
          
#ifdef INT_TIMER
          TimerReinstall();
#endif

          audiomute=0;

          /* continue emulation - end */
          /****************************/
          
          break;
        }
      } else 
        
#ifdef USE_BORDER_EFFECTS
        border_change();
#endif                                          
        /* Reset the display flags if required */
        if(LineNo==VmprChangeInLine-1) VmprChangeInLine=0;
        if(LineNo==ClutChangeInLine-1) ClutChangeInLine=0;
        if(LineNo==BorderChangeInLine-1) BorderChangeInLine=0;
    }
    
#ifdef USE_BORDER_EFFECTS
    if(BorderChangeInInstruction) {
      border_change();
      BorderChangeInInstruction=0;
    }
#endif
  }
}


void emulator_sync_timer(int sigparam) {
  /* Reset sync timer flag */
  SyncTimer=True;
}


void emulator_cancel_sync_timer(void) {
  //
  //Cancel the sync timer signal - called a closedown
  //
  static struct itimerval TmpTimer = { {0L, 0L}, {0L, 0L} };
  setitimer(ITIMER_REAL, &TmpTimer, NULL);
  signal(SIGALRM, SIG_IGN);
}

